/*!
 * \file agentThread.hpp
 * \brief This file is used to work on Jthread.
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 *
 * [License]
 *
 * $id$
 */

#ifndef AGENT_THREAD_HPP
#define AGENT_THREAD_HPP

#include <jni.h>
#include <jvmti.h>

/*!
 * \brief Entry point function type for jThread.
 * \param jvmti    [in] JVMTI environment information.
 * \param name     [in] Raw Monitor name.
 * \param userData [in] Pointer of user data.
 */
typedef void (JNICALL *TThreadEntryPoint)(jvmtiEnv *jvmti, JNIEnv *jni,
  void *userData);

/*!
* \brief This class is base class for parallel work using by jthread.
*/
class TAgentThread {
  public:
    /*!
     * \brief TAgentThread constructor.
     * \param name  [in] Thread name.
     */
    TAgentThread(const char *name);
    /*!
     * \brief TAgentThread destructor.
     */
    virtual ~TAgentThread(void);
    
    /*!
     * \brief Make and begin Jthread.
     * \param jvmti      [in] JVMTI environment object.
     * \param env        [in] JNI environment object.
     * \param entryPoint [in] Entry point for jThread.
     * \param args       [in] Pointer of user data.
     * \param prio       [in] Jthread priority.
     */
    virtual void start(jvmtiEnv *jvmti, JNIEnv *env,
      TThreadEntryPoint entryPoint, void *args, jint prio);
    /*!
     * \brief Notify timing to this thread from other thread.
     */
    virtual void notify(void);
    /*!
     * \brief Notify stopping to this thread from other thread.
     */
    virtual void stop(void);
    /*!
     * \brief Notify termination to this thread from other thread.
     */
    virtual void terminate(void);
    
  protected:
    /*!
     * \brief Number of request.
     */
    volatile int _numRequests;
    /*!
     * \brief Flag of exists termination request.
     */
    volatile bool _terminateRequest;
    /*!
     * \brief Flag of jThread flag.
     */
    volatile bool _isRunning;
    /*!
     * \brief Pthread mutex for thread.
     */
    pthread_mutex_t mutex;
    /*!
     * \brief Pthread mutex condition for thread.
     */
    pthread_cond_t mutexCond;
    /*!
     * \brief Thread name for java thread.
     */
    char *threadName;
};

#endif
