/*!
 * \file classContainer.hpp
 * \brief This file is used to add up using size every class.
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 *
 * [License]
 *
 * $id$
 */

#ifndef CLASS_CONTAINER_HPP
#define CLASS_CONTAINER_HPP

#include <limits.h>
#include <sys/time.h>
#include <string.h>
#include <jni.h>
#include <pthread.h>
#include <tr1/unordered_map>
#include <deque>

#include "util.hpp"
#include "trapSender.hpp"
#include "jvmInfo.hpp"
#include "sorter.hpp"
#include "snapShotContainer.hpp"

/*!
 * \brief This structure stored size of a class used in heap.
 */
typedef struct{
  jlong tag;   /*!< Pointer of TObjectData.                */
  jlong usage; /*!< Class using total size.                */
  jlong delta; /*!< Class delta size from before snapshot. */
} THeapDelta;

/*!
 * \brief This type is for map stored class information.
 */
typedef std::tr1::unordered_map<void*, TObjectData*,
                                TNumericalHasher<void *> > TClassMap;

/*!
 * \brief This type is for TClassContainer in Thread-Local-Storage.
 */
class TClassContainer;
typedef std::deque<TClassContainer*> TLocalClassContainer;

/*!
 * \brief This class is stored class information.<br>
 *        e.g. class-name, class instance count, size, etc...
 */
class TClassContainer {
  public:
    /*!
     * \brief TClassContainer constructor.
     */
    TClassContainer(TClassContainer *base = NULL, bool needToClr = true);
    /*!
     * \brief TClassContainer destructor.
     */
    virtual ~TClassContainer(void);
    
    /*!
     * \brief Append new-class to container.
     * \param klassOop [in] New class oop.
     * \return New-class data.
     */
    TObjectData *pushNewClass(void *klassOop);
    TObjectData *pushNewClass(void *klassOop, TObjectData *objData);
    
    /*!
     * \brief Remove class from container.
     * \param target [in] Remove class data.
     */
    void popClass(TObjectData *target);
    
    /*!
     * \brief Search class from container.
     * \param klassOop [in] Target class oop.
     * \return Class data of target class.
     */
    inline TObjectData *findClass(void *klassOop) {
      
      /* Search class data. */
      TClassMap::iterator it;
      bool notFound;

      SpinLockWait(&lockval);
      {
        it = classMap->find(klassOop);
        notFound = (it == classMap->end());
      }
      SpinLockRelease(&lockval);

      return notFound ? NULL : it->second;
    }
    
    /*!
     * \brief Update class oop.
     * \param oldKlassOop [in] Target old class oop.
     * \param newKlassOop [in] Target new class oop.
     * \return Class data of target class.
     */
    inline void updateClass(void *oldKlassOop, void *newKlassOop) {
      SpinLockWait(&lockval);
      {
        TClassMap::iterator it = classMap->find(oldKlassOop);

        if (it != classMap->end()) {
          /* Update class data. */
          it->second->klassOop   = newKlassOop;
          (*classMap)[newKlassOop] = it->second;
        
          /* Remove old klassOop. */
          classMap->erase(it);
        }

      }
      SpinLockRelease(&lockval);
    }

    /*!
     * \brief Get TLS ClassContainer
     */
    inline TClassContainer *getTLSContainer(void){
      TClassContainer *result = (TClassContainer *)pthread_getspecific(tlsKey);

      if(unlikely(result == NULL)){
        result = new TClassContainer(this, false);
        pthread_setspecific(tlsKey, result);

        SpinLockWait(&tlsLock);
        {
          classContainerInTLS.push_back(result);
        }
        SpinLockRelease(&tlsLock);

      }

      return result;
    }
    
    /*!
     * \brief Get class entries count.
     * \return Entries count of class information.
     */
    inline size_t getContainerSize(void) {
      size_t result;

      SpinLockWait(&lockval);
      {
        result = this->classMap->size();
      }
      SpinLockRelease(&lockval);

      return result;
    }
    
    /*!
     * \brief Remove all-class from container.
     */
    void allClear(void);
    /*!
     * \brief Output all-class information to file.
     * \param snapshot [in] Snapshot instance.
     * \return Sorted-class size information.
     */
    virtual TSorter<THeapDelta> *afterTakeSnapShot(TSnapShotContainer *snapshot);

  protected:
    /*!
     * \brief ClassContainer in TLS of each threads.
     */
    TLocalClassContainer classContainerInTLS;
    
  private:
    /*!
     * \brief SNMP trap sender.
     */
    TTrapSender *pSender;

    /*!
     * \brief Maps of class counting record.
     */
    TClassMap *classMap;
    
    /*!
     * \brief ClassContainer TLS key.
     */
    pthread_key_t tlsKey;

    /*!
     * \brief SpinLock variable.
     */
    int lockval;

    /*!
     * \brief SpinLock variable for TLS.
     */
    int tlsLock;

    /*!
     * \brief Do we need to clear at destructor?
     */
    bool needToClear;

};

#endif // CLASS_CONTAINER_HPP
