/*!
 * \file symbolFinder.cpp
 * \brief This file is used by search symbol in library.
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 *
 * [License]
 *
 * $id$
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <limits.h>
#include <link.h>
#include <bfd.h>
#include <errno.h>

#include "symbolFinder.hpp"
#include "util.hpp"

/*!
 * \brief Flag of BFD libarary is already initialized.
 */
bool TSymbolFinder::initFlag = false;

/*!
 * \brief Create record stored library information.
 * \param libinfo [out] Record of library information.
 * \param libPath [in]  Target library path.
 */
inline void createLibInfo(TLibraryInfo *libinfo, char const* libPath) {
  libinfo->libname_len = strlen(libPath);
  libinfo->libname = (char *)malloc(libinfo->libname_len + 1);
  
  /* If failure allocate memory. */
  if (unlikely(libinfo->libname == NULL)) {
    return;
  }
  
  strcpy(libinfo->libname, libPath);
}

/*!
 * \brief Callback function of dl_iterate_phdr(3).
 * \param info [in] Information of loaded library on memory.
 * \param size [in] Size of library area on memory.
 * \param data [in] User data.
 * \return Continues callback loop if value is 0.<br>
 *         Abort callback loop if value is 1.
 */
static int libraryCallback(struct dl_phdr_info *info, size_t size, void *data) {
  TLibraryInfo *libinfo = (TLibraryInfo *)data;
  
  /* If library is need. */
  if (strncmp(info->dlpi_name, libinfo->libname, libinfo->libname_len) == 0) {
    
    /* Clean old string. */
    free(libinfo->libname);
    
    /* Store library information. */
    createLibInfo(libinfo, info->dlpi_name);
    libinfo->baseaddr = info->dlpi_addr;
    
    /* Abort callback loop. */
    return 1;
  }
  
  return 0;
}

/*!
 * \brief TSymbolFinder constructor.
 */
TSymbolFinder::TSymbolFinder(void) {
  
  if (unlikely(!initFlag)) {
    initFlag = false;
    
    /* Global initialize. */
    bfd_init();
  }
  
  /* Zero clear records. */
  memset(&libBfdInfo, 0, sizeof(libBfdInfo));
  memset(&debugBfdInfo, 0, sizeof(debugBfdInfo));
  
  memset(&targetLibInfo, 0, sizeof(targetLibInfo));
}

/*!
 * \brief TSymbolFinder destructor.
 */
TSymbolFinder::~TSymbolFinder() {
  
  /* Cleanup. */
  free(targetLibInfo.libname);
  targetLibInfo.libname = NULL;
  
  if (likely(libBfdInfo.bfdInfo != NULL)) {
    bfd_close(libBfdInfo.bfdInfo);
    libBfdInfo.bfdInfo = NULL;
  }
  if (likely(debugBfdInfo.bfdInfo != NULL)) {
    bfd_close(debugBfdInfo.bfdInfo);
    debugBfdInfo.bfdInfo = NULL;
  }
  if (likely(libBfdInfo.staticSyms != NULL)) {
    free(libBfdInfo.staticSyms);
    libBfdInfo.staticSyms = NULL;
  }
  if (likely(debugBfdInfo.staticSyms != NULL)) {
    free(debugBfdInfo.staticSyms);
    debugBfdInfo.staticSyms = NULL;
  }
  if (likely(libBfdInfo.dynSyms != NULL)) {
    free(libBfdInfo.dynSyms);
    libBfdInfo.dynSyms = NULL;
  }
  if (likely(debugBfdInfo.dynSyms != NULL)) {
    free(debugBfdInfo.dynSyms);
    debugBfdInfo.dynSyms = NULL;
  }
}

/*!
 * \brief Load target libaray.
 * \param pathPattern [in] Pattern of target library path.
 * \return Is success load library.<br>
 */
bool TSymbolFinder::loadLibrary(const char *pathPattern) {
  /* Sanity check. */
  if (unlikely(pathPattern == NULL)) {
    PRINT_WARN_MSG("Failure allocate library path.");
    return false;
  }
  
  /* Initialize. */
  createLibInfo(&targetLibInfo, pathPattern);
  targetLibInfo.baseaddr = ((ptrdiff_t)0);
  
  /* If failure allocate memory. */
  if (unlikely(targetLibInfo.libname == NULL)) {
    PRINT_WARN_MSG("Failure allocate library path.");
    return false;
  }
  
  /* Search target library on memory. */
  if (unlikely(dl_iterate_phdr(&libraryCallback, &targetLibInfo) == 0)) {
    PRINT_WARN_MSG("Failure search library on memory.");
    free(targetLibInfo.libname);
    return false;
  }
  
  /* Load library with bfd record. */
  loadLibraryInfo(targetLibInfo.libname, &libBfdInfo);
  
  /* Create debuginfo path. */
  char dbgInfoPath[PATH_MAX + 1] = {0};
  snprintf(dbgInfoPath, PATH_MAX, DEBUGINFO_DIR "%s" DEBUGINFO_SUFFIX,
    targetLibInfo.libname);
  
  /* Load library's debuginfo with bfd record. */
  loadLibraryInfo(dbgInfoPath, &debugBfdInfo);
  
  /* If failure both load process. */
  if (unlikely(libBfdInfo.staticSymCnt == 0 && debugBfdInfo.staticSymCnt == 0
                  && libBfdInfo.dynSymCnt == 0 && debugBfdInfo.dynSymCnt == 0)) {
    
    PRINT_WARN_MSG("Failure load information about library.");
    free(targetLibInfo.libname);
    targetLibInfo.libname = NULL;
    return false;
  }
  
  return true;
}

/*!
 * \brief Find symbol in target library.
 * \param symbol [in] Symbol string.
 * \return Address of designated symbol.<br>
 *         value is null if process is failure.
 */
void *TSymbolFinder::findSymbol(char const* symbol) {
  void *result = NULL;
  
  /* Search symbol in library as static symbol. */
  
  if (likely(libBfdInfo.staticSymCnt != 0)) {
    result = doFindSymbol(&libBfdInfo, symbol, false);
  }
  
  if (unlikely(result == NULL)) {
    
    /* Search symbol in library's debuginfo. */
    
    if (likely(debugBfdInfo.staticSymCnt != 0)) {
      result = doFindSymbol(&debugBfdInfo, symbol, false);
    }
  }
  
  /* Search symbol in library as dynamic symbol. */
  
  if (unlikely(result == NULL)) {
    
    /* Search symbol in library. */
    
    if (likely(libBfdInfo.dynSymCnt != 0)) {
      result = doFindSymbol(&libBfdInfo, symbol, true);
    }
  }
  
  if (unlikely(result == NULL)) {
    
    /* Search symbol in library's debuginfo. */
    
    if (likely(debugBfdInfo.dynSymCnt != 0)) {
      result = doFindSymbol(&debugBfdInfo, symbol, true);
    }
  }
  
  if (likely(result != NULL)) {
    /* Convert absolute symbol address. */
    result = getAbsoluteAddress(result);
  }
  
  return result;
}

/*!
 * \brief Load target libaray information.
 * \param path  [in]  Target library path.
 * \param libInfo [out] BFD information record.<br>
 *        Value is null if process is failure.
 */
void TSymbolFinder::loadLibraryInfo(char const* path, TLibBFDInfo *libInfo) {
  bfd *bfdDesc = NULL;
  
  /* Open library. */
  bfdDesc = bfd_openr(path, NULL);
  if (unlikely(bfdDesc == NULL)) {
    return;
  }
  
  /* If illegal format. */
  if(!bfd_check_format(bfdDesc, bfd_object)) {
    bfd_close(bfdDesc);
    return;
  }
  
  /* If library don't have symbol. */
  if (!(bfd_get_file_flags(bfdDesc) & HAS_SYMS)) {
    bfd_close(bfdDesc);
    return;
  }
  
  asymbol *syms = NULL;
  
  /* Create empty symbol as temporary working space. */
  syms = bfd_make_empty_symbol(bfdDesc);
  if (unlikely(syms == NULL)) {
    bfd_close(bfdDesc);
    return;
  }
  
  /* Library is legal binary file. */
  libInfo->bfdInfo = bfdDesc;
  /* Load static symbols. */
  libInfo->staticSymCnt = bfd_read_minisymbols(bfdDesc, 0,
                    (void **)&libInfo->staticSyms, &libInfo->staticSymSize);
  /* Load dynamic symbols. */
  libInfo->dynSymCnt = bfd_read_minisymbols(bfdDesc, 1,
                          (void **)&libInfo->dynSyms, &libInfo->dynSymSize);
  libInfo->workSym = syms;
}

/*!
 * \brief Find symbol in target library to use BFD.
 * \param libInfo  [in] BFD information record.
 * \param symbol   [in] Symbol string.
 * \param isDynSym [in] Symbol is dynamic symbol.
 * \return Address of designated symbol.<br>
 *         value is null if process is failure.
 */
void *TSymbolFinder::doFindSymbol(TLibBFDInfo *libInfo,
                                     char const* symbol, bool isDynSym) {
  
  bfd *bfdDesc        = libInfo->bfdInfo;
  asymbol *syms       = libInfo->workSym;
  char *miniSymsEntry = NULL;
  int miniSymCnt      = 0;
  int dyn             = ((isDynSym) ? 1 : 0);
  void *result        = NULL;
  unsigned int size   = 0;
  
  /* Load symbol list from binary file. */
  if (isDynSym) {
    miniSymCnt    = libInfo->dynSymCnt;
    miniSymsEntry = libInfo->dynSyms;
    size          = libInfo->dynSymSize;
  } else {
    miniSymCnt    = libInfo->staticSymCnt;
    miniSymsEntry = libInfo->staticSyms;
    size          = libInfo->staticSymSize;
  }
  
  /* Search symbol. */
  for (int idx = 0; idx < miniSymCnt; idx++) {
    
    /* Convert symbol to normally symbol. */
    asymbol *symEntry = bfd_minisymbol_to_symbol(bfdDesc, dyn,
                                                    miniSymsEntry, syms);
    
    if (strcmp(bfd_asymbol_name(symEntry), symbol) == 0) {
      
      /* Found target symbol. */
      result = (void *)bfd_asymbol_value(symEntry);
      break;
    }
    
    /* Move next symbol entry. */
    miniSymsEntry += (ptrdiff_t)size;
  }
  
  return result;
}

